***********************************************************************;
*  AdaBoost:  A macro to perform boosting using stumps--recursive     *;
*             partitions with one split.                              *;
*                                                                     *;
*  Inputs:                                                            *;
*  inputds:      name of the input data set.  This data set must      *;
*                have descriptors labeled x1, x2,..., xp.  The        *;
*                response must be labeled as y and must have values   *;
*                0 and 1.                                             *;
*  p:            the number of descriptors in the input data set.     *;
*  iter:         the number of boosting iterations desired.           *;
*                                                                     *;
*  Output:                                                            *;
*  boost:        a data set that contains information about each      *;
*                iteration from boosting.  The data set contains the  *;
*                following variables:                                 *;
*                iter:  boosting iteration number                     *;
*                gini_var:  the descriptor number at the current      *;
*                            iteration that minimizes the Gini Index  *;
*                gini_cut:  the cut-point that minimizes Gini         *;
*                p0_LH:   the probability of class 0 for observations *;
*                         less than gini_cut                          *;
*                p1_LH:   the probability of class 1 for observations *;
*                         less than gini_cut                          *;
*                C_L:     the class label for observations less than  *;
*                         gini_cut                                    *;
*                p0_RH:   the probability of class 0 for observations *;
*                         greater than gini_cut                       *;
*                p1_RH:   the probability of class 1 for observations *;
*                         greater than gini_cut                       *;
*                C_R:     the class label for observations greater    *;
*                         than gini_cut                               *;
*                alpha:   the weight of the SPLIT at the current      *;
*                         boosting iteration                          *;
*                error:   the misclassification error of the          *;
*                         cumulative boosting model at the current    *;
*                         iteration                                   *;
*                kappa:   the kappa statistic of the cumulative       *;
*                         boosting model at the current iteration     *;
***********************************************************************;

%macro AdaBoost(inputds,p,iter);
  proc iml;
    ****************************************************************;
    *  IML_SPLIT:  An IML module to find the optimal binary split  *;
    *              using the Gini Index                            *;
    *  Inputs:                                                     *;
    *    x:    an n x p matrix of the descriptors                  *;
    *    y:    an n x 1 vector of the response, coded as 0 and 1   *;
    *    w:    an n x 1 vector of the observation weights          *;
    *                                                              *;
    *  Outputs:                                                    *;
    *    g_info:    a 1 x 8 vector that contains the following     *;
    *               information:                                   *;
    *               gini_var:   the variable number that minimizes *;
    *                           the Gini Index                     *;
    *               gini_cut:   the cut-point that minimizes the   *;
    *                           Gini Index                         *;
    *               p0_LH:      the probability of class 0 for     *;
    *                           observations less than gini_cut    *;
    *               p1_LH:      the probability of class 1 for     *;
    *                           observations less than gini_cut    *;
    *               C_L:        the class label for observations   *;
    *                           less than gini_cut                 *;
    *               p0_RH:      the probability of class 0 for     *;
    *                           observations greater than gini_cut *;
    *               p1_RH:      the probability of class 1 for     *;
    *                           observations greater than gini_cut *;
    *               C_R:        the class label for observations   *;
    *                           greater than gini_cut              *;
    *    y_pred:    an n x 1 vector of predicted classification    *;
    *               for each observation                           *;
    ****************************************************************;
    start iml_split(x,y,w,g_info,y_pred);
        n = nrow(x);
        p = ncol(x);
        gini_min = 2;
        gini_var = 0;
        gini_cut = 0;
        y_pred = repeat(0,n,1);
        do j=1 to p;
            x_curr = x[,j]||y||w;
            b=x_curr;
            x_curr[rank(x[,j]),]=b;
            x_sort = x_curr[,1];
            y_sort = x_curr[,2];
            w_sort = x_curr[,3];
            L = (repeat((1:n-1),n,1) >= repeat((1:n)`,1,n-1));
            R = (repeat((1:n-1),n,1) < repeat((1:n)`,1,n-1));
            p1_L = (y_sort#w_sort#L)[+,]#(1/(w_sort#L)[+,]);
            p0_L = ((1-y_sort)#w_sort#L)[+,]#(1/(w_sort#L)[+,]);
            p1_R = (y_sort#w_sort#R)[+,]#(1/(w_sort#R)[+,]);
            p0_R = ((1-y_sort)#w_sort#R)[+,]#(1/(w_sort#R)[+,]);
            lpwt = (w_sort#L)[+,];
            rpwt = (w_sort#R)[+,];
            gini = p1_L#p0_L#lpwt + p1_R#p0_R#rpwt;
            g_ord = gini;
            g_ord[,rank(g_ord)]=(1:n-1);
            if gini[g_ord[1,1]] < gini_min then do;
            gini_min=gini[g_ord[1,1]];
            gini_var = j;
            gini_cut = (x_sort[g_ord[1,1]] + x_sort[g_ord[1,1]+1]) / 2;
            p1_RH = p1_R[g_ord[1,1]];
            p0_RH = p0_R[g_ord[1,1]];
            p1_LH = p1_L[g_ord[1,1]];
            p0_LH = p0_L[g_ord[1,1]];
            c_R = 0;
            if p1_RH > 0.5 then c_R = 1;
            c_L = 0;
            if p1_LH > 0.5 then c_L = 1;
            end;
        end;
        g_info = gini_var||gini_cut||p0_LH||p1_LH||c_L||p0_RH||p1_RH||c_R;
        do i=1 to n;
            if x[i,gini_var] <= gini_cut then y_pred[i,1] = c_L;
            if x[i,gini_var] >  gini_cut then y_pred[i,1] = c_R;
        end;
    finish iml_split;

    start kappa(y,y_pred,kappa);
        A = sum(y_pred=0) - sum(y>y_pred);
        B = sum(y_pred>y);
        C = sum(y>y_pred);
        D = sum(y_pred=1) - sum(y_pred>y);
        N = (A+B+C+D);
        O = (A+D)/N;
        E = ((A+C)*(A+B)+(B+D)*(C+D))/(N*N);
        if E = 1 then kappa = 1;
        else kappa = (O-E)/(1-E);
    finish kappa;

    * 0. Import data;
    use &inputds;
        read all var('x1':"x&p") into x;
        read all var {y} into y;
    * AdaBoost;
    * 1.  Initialize observation weights and overall prediction;
    w = repeat(1/nrow(y),nrow(y),1);
    sum_pred = repeat(0,nrow(y),1);
    * 2.  For t=1 to T do;
    %do t=1 %to &iter;
        * 2.a.  Fit model;
        run iml_split(x,y,w,g_info,y_pred_t);
        * 2.b.  Compute contribution of each observation;
        err_t  = (w#(y_pred_t ^= y))[+,];
        if err_t = 0 then err_t = 0.0001;
        if err_t = 1 then err_t = 0.9999;
        alpha_t = log((1-err_t)/err_t);
        c_t = alpha_t*(2*y_pred_t-1);
        sum_pred = sum_pred + c_t;
        * 2.c.  Update weights;
        w = (w#exp(alpha_t*(y_pred_t^=y))/sum(w#exp(alpha_t*(y_pred_t^=y))));
        * Accuracy of classifier at current stage;
        y_pred = (sum_pred > 0);
        error = (sum(y ^= y_pred))/nrow(y);
        run kappa(y,y_pred,kappa);
        iter = &t;
        ada_t = iter||g_info||alpha_t||error||kappa;
        iterinfo = iterinfo//ada_t;
    %end;
    * 3.  Compute final classifier;
    ada_pred = (sum_pred > 0);
    * Output boosting information;
    varname1 = {'iter' 'gini_var' 'gini_cut' 'p0_LH' 'p1_LH' 'c_L' 'p0_RH'
                'p1_RH' 'c_R' 'alpha' 'error' 'kappa'};
    create boost from iterinfo [colname=varname1];
    append from iterinfo;
    quit;
%mend AdaBoost;