***********************************************************************;
*  SPLIT:  A macro to find the optimal binary split                   *;
*          using the Gini Index                                       *;
*  Inputs:                                                            *;
*  nputds:     the input data set                                     *;
*              must contain descriptors named x1, x2, ..., xp,        *;
*              a variable named w that provides each                  *;
*              observation weight, and the response named y           *;
*              that takes values 0 and 1 for each class               *;
*  p:          the number of descriptors                              *;
*  Output:                                                            *;
*              The macro generates an output data set named           *;
*              OUTSPLIT which contains the name of the variable       *;
*              and split point that minimizes the Gini Index.         *;
*              OUTSPLIT also contains the value of the Gini Index     *;
***********************************************************************;
%macro split(inputds,p);
    * Get number of observations;
    proc sql noprint;
        select count(*)
        into : n
        from &inputds;
        quit;
    %let gini_min = 2;
    %let gini_var = 0;
    %let gini_cut = 0;
    %do j=1 %to &p;
        proc sort data=&inputds(keep=x&j y w) out=sorted;
        by x&j;
        run;
        %let finalobs = %sysevalf(&n-1);
        %do i=1 %to &finalobs;
            proc means data=sorted(obs=&i) mean noprint;
            var y;
            weight w;
            output out=p1_1(drop=_TYPE_ _FREQ_) mean=p1_1;
            run;
            %let next_obs = %sysevalf(&i+1);
            proc means data=sorted(firstobs=&next_obs) mean noprint;
                var y;
                weight w;
                output out=p1_2(drop=_TYPE_ _FREQ_) mean=p1_2;
                run;    
            data _null_;
                merge p1_1 p1_2;
                ppn1 = &i/&n;
                ppn2 = 1-ppn1;
                gini = 2*p1_1*(1-p1_1)*ppn1 + 2*p1_2*(1-p1_2)*ppn2;
                call symput('gini',left(gini));
                run;
            %if &gini < &gini_min %then %do;
                %let gini_min = &gini;
                %let best_j = &j;
                proc sql noprint;
                    select mean(x&j)
                    into : cutoff
                    from sorted(firstobs=&i obs=&next_obs);
                    quit;
            %end;
        %end;
    %end;
    data outsplit;
        gini=&gini_min;
        best_var="x&best_j";
        cutoff=&cutoff;
        label gini="Gini index"
              best_var="Best variable"
              cutoff="Best cutoff"; 
        run;
%mend split;
