%macro dscf(dataset,group,var,alpha);
***************************************************************;
* The Dwass Steele Critchlow-Fligner Macro                    *;
* ------------------------------------------------------------*;
*                                                             *;
* This macro is designed to perform all two-sided pair-wise   *;
* comparisons of the location parameters for n groups (n gt 2)*;
* in a one-way layout. The method for the procedure involves a*;
* large sample approximation. The method of comparison employs*;
* pairwise Wilcoxon statistics for all pairs of groups in the *;
* the data. The macro also contain an adjustment for ties.    *;
*                                                             *;
* The required inputs for the macro are a SAS data set        *; 
* (DATASET) containing a character class variable (GROUP),    *;
* a response variable (VAR), and a family-wise error rate     *;
* (ALPHA).                                                    *;
*                                                             *;
* For more details on this nonparametric multiple comparison  *;
* procedure, see Hollander and Wolfe s Nonparametric          *;
* Statistical Methods, 2/e pp. 240-248.                       *;
*                                                             *;
* Overall Macro Scheme:                                       *;
* --------------------                                        *;
* The macro consists of a body of code containing one embedded*;
* macro (%GROUPS). The embedded macro determines the number of*;
* groups present (NGRPS). If a group in the SAS data set does*;
* not contain at least one response value, it will not be     *;
* included in the analysis. The embedded macro also creates   *;
* two global macro variables that contain the group labels    *;
* (GRPVEC) for the levels of the class variable and          *;
* information about the sample size for each group (NVEC).   *;
*                                                             *;
* The main body of the SAS macro code determines summary      *;
* statistics (e.g., Wilcoxon test statistics) and the number  *;
* of ties present in the pair-wise comparison. This           *;
* information is then employed to calculate the pair-wise     *;
* test statistics. The cutoff for the test statistic is       *;
* calculated with PROBMC function. As the macro iterates      *;
* between all pair-wise comparisons it concantenates          *;
* successive SAS data sets in a data set called STAT. The     *;
* results are then printed out with a PROC PRINT procedure.   *;
*                                                             *;
***************************************************************;

%macro groups;
* This macro creates three outputs:                        *;
* (1) the total number of groups present (&ngrps)          *; 
* (2) a macro variable that contains all of the individual *;
*     class variable sample sizes (&nvec)                  *;
* (3) a macro variable that contains all of the individual *;
*     class variable labels (&grpvec)                      *;
%global grpvec nvec;
* remove missing response values;
data &dataset.2;
    set &dataset;
    if &var ne .;
proc sort data=&dataset.2; 
    by &group;
* Identify levels of class variable;
data &group; 
    set &dataset.2; 
    by &group;
    if first.&group;
proc transpose data=&group out=grpvec; 
    var &group;
* Determine the total sample size for the experiment;
data _null_;
    call symput('ngrps',left(put(count,8.)));
    stop;
    set &group nobs=count;
    run;
* Determine the sample size for each level of the class variable (group);
proc univariate noprint data=&dataset.2;
    var &var;
    by &group;
    output out=n&var n=n&var;
proc transpose data=n&var out=nvec; 
    var n&var;
* Create a macro variable that contains each class level s sample size for later processing;
data _null_;
    set nvec;
    nvec=
    %do s=1 %to %eval(&ngrps-1);
        trim(left(put(col&s,8.)))||" "||
    %end;
    trim(left(put(col&ngrps,8.)));
    call symput('nvec',nvec);
    run;
* Create a macro variable that contains a label for each level of the class variable;
data _null_; 
    set grpvec;
    grpvec=
    %do g=1 %to %eval(&ngrps-1);
        trim(left(col&g))||" "||
    %end;
    trim(left(col&ngrps));
    call symput('grpvec',grpvec);
    run;
%mend groups;
%groups
* Start iterative processing with an empty data sas data set called stat;
data stat;
* Start calculating pair-wise statistics necessary for the multiple comparison procedure;
    %do u=1 %to &ngrps;
        %do v=%eval(&u+1) %to &ngrps;
        * Determine sample sizes of two groups involved in comparisons;
            proc means data=&dataset.2 noprint n;
                by &group;
                where &group in ("%scan(&grpvec,&u," ")","%scan(&grpvec,&v," ")");
                var &var;
                output out=n&u&v(keep=n) n=n;
        * Calculate pair-wise wilcoxon statistics for group u and group v;
            proc npar1way data=&dataset.2 wilcoxon noprint;
                where &group in ("%scan(&grpvec,&u," ")","%scan(&grpvec,&v," ")");
                class &group;
                var &var;
                output out=w&u&v(keep=_wil_);
        * Determine the number of ties in the original data set (with all missing values removed;
            proc freq data=&dataset.2 noprint;
                where &group in ("%scan(&grpvec,&u," ")","%scan(&grpvec,&v," ")");
                tables &var/out=t&u&v(keep=count);
                run;
        * Calculate tie-adjustment factor necessary for variance estimate in test statistic;
            data t&u&v;
                set t&u&v;
                tie=(count-1)*(count)*(count+1);
                run;
        * Calculate the sum value of tie-adjustment value;
            proc means data=t&u&v sum noprint;
                var tie;
                output out=tsum(keep=tiesum) sum=tiesum;
            proc transpose data=n&u&v out=n1n2(keep=n1 n2) prefix=n;
                var n;
            data n1n2;
                set n1n2;
                mergeid+1;
            data w&u&v;
                set w&u&v;
                mergeid+1;
            data tsum;
                set tsum;
                mergeid+1;
            proc sort data=n1n2;
                by mergeid;
            proc sort data=w&u&v;
                by mergeid;    
            proc sort data=tsum;
                by mergeid;
        * Put together wilcoxon statistics sample sizes and tie information together
          for test statistic calculation. calculate test statistic;
            data s&u&v;
                merge w&u&v n1n2 tsum;
                by mergeid;
                drop mergeid;
                length difflbl $50;
                nsum= n1 + n2;
                min_n=min(n1,n2);
                numer1= _wil_;
                numer2= (min_n*(nsum+1))/2;
                numer= numer2-numer1;
                var= ((n1*n2)/24)*(nsum+1-(tiesum/(nsum*(nsum-1))));
                denom=sqrt(var);
                w=(numer/denom); * Test statistic for dwass steele critchlow-fligner statistics;
                absw=abs(w); * Absolute value used for two-sided testing;
                qalpha = probmc('range',.,(1-&alpha),.,&ngrps); %*studentized-range statistics cutoff;
                if absw ge qalpha then do;
                    symbol="**"; * Statistical significance indicator;
                end;
                difflbl="%scan(&grpvec,&u," ") - %scan(&grpvec,&v," ")";
            data stat;
                set stat s&u&v;
                run; 
         %end;
    %end;
data stat;
    set stat;
    where n1 ne . and n2 ne .; * Remove original record from the empty data stat;
    c+1;
    label difflbl="Group@comparisons"
          w="Test@statistic"
    	  absw="Test statistic@absolute@value"
          qalpha="Cutoff@at@alpha=&alpha" 
    	  symbol="Significant@difference = **"
          c="Comparison@number"
    ;
    run;
    title1 "Large sample approximation multiple comparison procedure";
    title2 "or all treatment pairs";
    title3 "based upon pairwise rankings";
    title4 "%trim(%left(&ngrps)) groups: %trim(%left(&grpvec)) (respective sample sizes: %trim(%left(&nvec)))";
    title5 "Alpha = &alpha";    

* Print out test statistic results from all pair-wise comparisons;
proc print data=stat noobs split="@";
    var difflbl c w absw qalpha symbol;
    run;
%mend dscf;
