%macro dunn(dataset,group,var,alpha);
***************************************************************;
* The Dunn Macro                                              *;
* ------------------------------------------------------------*;
*                                                             *;
* This macro is designed to perform all two-sided pair-wise   *;
* comparisons of the location parameters for n groups (n gt 2)*;
* in a one-way layout. The method for the procedure involves a*;
* large sample approximation. The method of comparison employs*;
* ranking all non-missing observations and averaging the ranks*;
* for each level of the class variable (GROUP).               *;
*                                                             *;
* The required inputs for the macro are a SAS data set        *;
* (DATASET) containing a character class variable (GROUP),    *;
* a response variable (VAR) and a family-wise error rate      *;
* (ALPHA).                                                    *;
*                                                             *;
* For more details on this nonparametric multiple comparison  *;
* procedure, see Hollander and Wolfe s Nonparametric          *;
* Statistical Methods, 1/e p. 125.                            *;
*                                                             *;
* Overall Macro Scheme:                                       *;
* --------------------                                        *;
* The macro consists of a body of code containing one embedded*;
* macro (GROUPS). The embedded macro determines the number of*;
* groups present (NGRPS). If a group in the SAS data set does *;
* not contain at least one response value, it will not be     *;
* included in the analysis. The embedded macro also creates   *;
* one global macro variable that contains the group labels    *;
* (GRPVEC) for the levels of the class variable.              *;
*                                                             *;
* The main body of the SAS macro code determines summary      *;
* statistics (e.g., average ranks, sample sizes, etc.) This   *;
* information is then employed to calculate the pair-wise     *;
* test statistics. The cutoff for the test statistic is       *;
* calculated with PROBIT function. The results are then       *;
* printed out with a PROC PRINT procedure.                    *;
*                                                             *;
***************************************************************;

* First, take the input data set &dataset and eliminate missing values;
data &dataset.2;
    set &dataset;
    if &var ne .;
    run;

%macro groups;
* This macro creates two outputs:                         *;
* (1) The total number of groups present (NGRPS)          *; 
* (2) A macro variable that contains all of the individual*;
*     class variable labels (GRPVEC)                      *;
%global grpvec;
proc sort data=&dataset.2; 
    by &group;
* Create a data set that contains just the levels of the class variable (group);
data &group; 
    set &dataset.2; 
    by &group;
    if first.&group;
proc transpose data=&group out=grpvec; 
    var &group;
* Determine the number of levels of the class variable;
data _null_;
    call symput('ngrps',left(put(count,8.)));
    stop;
    set &group nobs=count;
    run;
* Create a global macro variable containing the labels for each level of the class variable;
data null; 
    set grpvec;
    grpvec=
    %do g=1 %to %eval(&ngrps-1);
        trim(left(col&g))||" "||
    %end;
    trim(left(col&ngrps));
    call symput('grpvec',grpvec);
    run;
%mend groups;
%groups;
proc sort data=&dataset.2;
    by &group;
* Rank all non-missing responses;
proc rank data=&dataset.2 out=r&var;
    ranks r&var;
    var &var;
proc sort data=r&var;
    by &group;
* Calculate the average of the ranks for each group;
proc univariate noprint data=r&var;
    var r&var;
    by &group;
    output out=s&var sum=s&var n=n&var;
* Determine the total sample size for the entire experiment;
proc univariate noprint data=s&var;
    var n&var;
    output out=sn&var sum=sn&var;
*C reate a macro variable with value equal to the total sample size for the entire experiment;
data _null_;
    set sn&var;
    call symput("n",left(put(sn&var,8.)));
    run;
data sr&var; 
    set s&var;
    keep s&var;
data n&var; 
    set s&var;
    keep n&var;
proc transpose data=sr&var out=srvec; 
    var s&var;
* Create a macro variable containing the summary statistics (average ranks) for each group;
data _null_;
    set srvec;
    srvec=
    %do s=1 %to %eval(&ngrps-1);
        trim(left(put(col&s,11.3)))||" "||
    %end;
    trim(left(put(col&ngrps,11.3)));
    call symput('srvec',srvec);
    run;
proc transpose data=n&var out=nvec; 
    var n&var;
* Create a macro variable containing the sample sizes for each level of the class variable;
data _null_;
    set nvec;
    nvec=
    %do s2=1 %to %eval(&ngrps-1);
        trim(left(put(col&s2,8.)))||" "||
    %end;
    trim(left(put(col&ngrps,8.)));
    call symput('nvec',nvec);
    run;
* Perform the pair-wise large sample approximation mcp;
data nparmcp;
    %do u=1 %to &ngrps;
        %do v=%eval(&u+1) %to &ngrps;
            %let rsum&u=%scan(&srvec,&u," ");
            %let n&u=%scan(&nvec,&u," ");
            %let rsum&v=%scan(&srvec,&v," ");
            %let n&v=%scan(&nvec,&v," ");
            c+1;
            difflbl="%scan(&grpvec,&u," ")-%scan(&grpvec,&v," ")";
            avdiff=abs((&&rsum&u/&&n&u)-(&&rsum&v/&&n&v));
            cutoff=probit(1-(&alpha/((&ngrps*(&ngrps-1)))))*(((&n*(&n+1))/12)**0.5)*((1/&&n&u + 1/&&n&v)**0.5);
            symbol="  ";
            if avdiff>=cutoff then do;
                symbol="**";
            end;
         output;
         %end;
    %end;
    label difflbl="Group@comparisons"
          avdiff="Difference@in@average@ranks"
          cutoff="Cutoff@at@alpha=&alpha"
          symbol="Significance@difference = **"
          c="Comparison@number"
    ;
    title1 "Large sample approximation multiple comparison procedure";
    title2 "designed for unbalanced data";
    title3 "%trim(%left(&ngrps)) groups: %trim(%left(&grpvec)) (respective sample sizes: %trim(%left(&nvec)))";
    title4 "Alpha = &alpha";
* Print out the results of the test;
proc print label split='@' noobs;
    run;
%mend dunn;

