%macro fpstatistic(dataset,var,groupvar);
%***************************************************************;
%* A macro for performing a nonparametric comparison of the    *;
%* location parameter between two independent groups in the    *;
%* presence of unequal scale.                                  *;
%* ------------------------------------------------------------*;
%*                                                             *;
%* This macro is designed to perform the Fligner-Policello     *;
%* test, a nonparametric approach to the Behrens-Fisher        *;
%* problem (a setting where one wishes to compare two location *;
%* parameters from two distributions with unequal variation).  *;
%* For more information, see Fligner, M.A., Policello, G.E.    *;
%* (1981) Robust rank procedures for the Behrens-Fisher        *; 
%* problem. JASA. 76, 162-174.                                 *;
%*                                                             *;
%* The required inputs for the macro are a SAS data set        *;
%* (dataset) containing a character class variable             *;
%* (groupvar) and a response variable (var).                   *;
%*                                                             *;
%* Overall macro scheme:                                       *;
%* --------------------                                        *;
%* The first portion of the macro is devoted to calculating the*;
%* placements for both samples. to calculate the placements,   *;
%* the macro first determines the two groups respective sample *;
%* sizes and then loops through both data sets to determine    *;
%* the full set of placements for each group. the form of the  *;
%* placement determination implemented in this macro includes  *;
%* a correction for tied values.                               *;
%*                                                             *;
%* The second half of the program calculates the average       *;
%* placements and variation of the placements necessary to     *;
%* for the fligner-policello statistic. the last portion of the*;
%* sas code performs the large sample approximate test and     *;
%* provides the corresponding p-values.  the results are then  *;
%* printed out with proc print.                                *;
%*                                                             *;
%* Note: as programmed here, this analysis can take a few      *;
%* minutes to run for large data sets.                         *;
%*                                                             *;
%***************************************************************;

* Sort data by classification or grouping variable;
proc sort data=&dataset;
    by &groupvar;
* Determine group sample sizes;
proc means data=&dataset noprint;
    var &var;
    by &groupvar;
    output out=placen n=n;
* Output, as a macro variable, the sample size of the first;
* Level of the classification variable as determined by sort order;
data _null_;
    set placen;
    if _n_=1;
    call symput("n1",left(put(n,8.)));
    run;
* Output, as a macro variable, the sample size of the second;
* level of the classification variable as determined by sort order;
data _null_;
    set placen;
    if _n_=2;
    call symput("n2",left(put(n,8.)));
    run;
* Set up an empty sas data set to store iterative placement determinations;
* for first group in sort order;
data p_fpstatistic;
* Determine placements for first group in sort order;
%do i=1 %to &n1;
data p&i;
    set &dataset;
    if _n_=&i or _n_ > &n1;
    if _n_=&i then do; 
        x&i=&var;
        retain x&i;
    end;
    if _n_>=&n1+1;
        if &var < x&i then do;
            pind&i=1; * Assign a value of 1 for values less than first sample;
        end;
    else if &var = x&i then do;
        pind&i=0.5; * Assign a value of 1/2 for tied values;
    end;
    else do;
        pind&i=0;
    end;
* Determine placement values by summing up indicators;
proc means data=p&i noprint;
    var pind&i;
    output out=p&i(keep=p&i) sum=p&i;
data p&i;
    set p&i;
    rename p&i=p; 
* Iteratively add placements to original empty dataset;
* Remove first empty record;
data p_fpstatistic;
    set p_fpstatistic p&i;
    if p>.;
    run;
%end; * End of n1 loop;
* Set up an empty sas data set to store iterative placement determinations;
* for second group in sort order;
data q_fpstatistic;
* Determine placements for second group in sort order;
%do j=1 %to &n2;
proc sort data=&dataset;
    by descending &groupvar;
data q&j;
    set &dataset;
    if _n_=&j or _n_ > &n2;
    if _n_ = &j then do; 
        y&j=&var;
        retain y&j;
    end;
    if _n_ >= &n2+1;
        if &var < y&j then do;
            qind&j=1; %*assign a value of 1 for values less than second sample;
        end;
        else if &var = y&j then do;
            qind&j=0.5; %*assign a value of 1/2 for tied values;
        end;
    else do;
        qind&j=0;
    end;
* Determine placement values by summing up indicators;
proc means data=q&j noprint;
    var qind&j;
    output out=q&j(keep=q&j) sum=q&j;
data q&j;
    set q&j;
    rename q&j=q;
* Iteratively add placements to original empty dataset;
* Remove first empty record;
data q_fpstatistic;
    set q_fpstatistic q&j;
    if q>.;
    run;
%end; * End of n2 loop;
* Calculate placement statistics for the first group;
proc means data=p_fpstatistic noprint;
    var p;
    output out=pstats sum=sumofpis mean=pbar css=v1;
* Create macro variables for each of the first groups placement statistics;
data _null_;
    set pstats;
    call symput("sumofpis",left(put(sumofpis,15.7))); 
    call symput("pbar",left(put(pbar,15.7)));
    call symput("v1",left(put(v1,15.7))); 
    run;
* Calculate placement statistics for the second group;
proc means data=q_fpstatistic noprint;
    var q;
    output out=qstats sum=sumofqjs mean=qbar css=v2;
* Create macro variables for each of the second groups placement statistics;
data _null_;
    set qstats;
    call symput("sumofqjs",left(put(sumofqjs,15.7))); 
    call symput("qbar",left(put(qbar,15.7)));
    call symput("v2",left(put(v2,15.7))); 
    run;
* Calculate the statistic and corresponding p-values;
* second level of grouping or classification variable compared with first level;
* i.e., test is second level of class variable - first level of class variable;
data fpstatistic;
    n1=&n1;
    n2=&n2;
    uhattop=&sumofqjs-&sumofpis;
    uhatbottom=(&v1+&v2+(&pbar*&qbar))**0.5;
    uhat=uhattop/(2*uhatbottom);
    uhat2sd=abs(uhat);
    p2sided=2*(1-probnorm(uhat2sd));
    up1sided=1-probnorm(uhat);
    low1sided=probnorm(uhat);
    keep n1 n2 uhat p2sided up1sided low1sided;
    label n1="Sample*size*group 1"
          n2="Sample*size*group 2"
    	  uhat="FP*statistic"
    	  p2sided="Two-sided*p-value"
    	  up1sided="Upper-tailed*p-value"
    	  low1sided="Lower-tailed*p-value";
    run;
proc print data=fpstatistic noobs label split="*";
    var n1 n2 uhat p2sided up1sided low1sided;
    run;
%mend fpstatistic;