%macro kwss(dataset,group,var,shiftlist,sample_size,iteration_no,alpha);
*******************************************************************************;
* Kruskall-Wallis Sample Size Determination Macro                             *;
* ----------------------------------------------------------------------------*;
*                                                                             *;
* This macro was designed to perform the Mahoney-Magel generalization of the  *;
* Collings-Hamilton approach for estimating the power of a comparison in the  *;
* presence of a location shift.  The macro achieves the estimation of power   *;
* via a resampling scheme AND IS VERY RESOURCE INTENSIVE.                     *;
*                                                                             *;
* The required inputs of the macro are a SAS data set (DAASET) containing a   *;
* character grouping variable (GROUP) and a response variable (VAR). The      *;
* user must also specify a list of "shifts" or changes in location            *;
* (SHIFTLIST), the desired sample size for power determination                *;
* (SAMPLE_SIZE), the number of interations for resampling with replacement    *;
* (ITERATION_NO) and the level of statistical significance for the test       *;
* (ALPHA).                                                                    *;
*                                                                             *;
* For more details on the statistical theory of this technique, please see    *;
* "Estimating the Power of the Two-Sample Wilcoxon Test for Location Shift",  *;
* B.J. Collings & M.A. Hamilton, Biometrics 44, pp. 847-860 and "Estimation of*;
* the Power of the Kruskal-Wallis Test", M. Mahoney & R. Magel, Biometrical   *;
* Journal, 38, pp. 613-630.                                                   *;
*                                                                             *;
* Overall Macro Scheme:                                                       *;
* --------------------                                                        *;
* The macro consists of a body of code containing one embedded macro          *;
* (%BOOTSTRAP). The program begins by removing all observations in the SAS    *;
* input data set (DATASET) that contain missing values for the response (VAR).*;
* The macro then determines the number of levels of the classification or     *;
* grouping variable (VAR) and records that value as a macro variable NGRPS.   *;
* The macro then creates a macro variable (GRPVEC) that contains a list of    *;
* all levels of the classification or grouping variable (GROUP). The program  *;
* determines the number of location shifts specified in SHIFTLIST.  This      *;
* value is compared with the number of groups (NGRPS).  If the two do not     *;
* agree, the program terminates. For each level of GRPVEC, the program        *;
* resamples DATASET SAMPLE_SIZE*NGRPS times, and repeats the procedure        *;
* ITERATION_NO times.  E.g., if one is interested in comparing 5 groups       *;
* of size 7 for 5 prespecified shifts in location at level ALPHA, the         *;
* program will begin by creating a bootstrapped sample at the first level of  *;
* the grouping variable (GROUP) of size 7*5=35 observations.  The program     *;
* then applies the 5 shifts to the 35 observations and employs PROC NPAR1WAY  *;
* to test for a shift in location.  If the result is statistically significant*;
* at level ALPHA, an indicator variable (SIGCOUNT) is assigned the value "1". *;
* This procedure is repeated a large number of times (ITERATION_NO, typically *;
* > 1000). At the conclusion, the proportion of "1's" is determined for the   *;
* first level of the grouping variable. The above procedure is then repeated  *;
* for the remaining 4 levels of the grouping variable.  The process concludes *;
* by using proc means to determine the mean power for all 5 groups and reports*;
* its value with PROC PRINT.  The titles before the PROC PRINT recite         *;
* information about the execution of the calculations.                        *;
*                                                                             *;
*******************************************************************************;

data POWER;
 PERCENT=-0.99;
run;

%*First, take the input data set &dataset and eliminate missing values;
data &dataset.2;
 set &dataset;
 if &VAR ne .;
run;

%*Determine the number of levels of the class variable;
proc sort data=&dataset.2; by &GROUP;

data &GROUP; set &dataset.2; by &GROUP;
if first.&GROUP;
run;

data _null_;
call symput('NGRPS',left(put(count,8.)));
stop;
set &GROUP nobs=COUNT;
run;

proc transpose data=&GROUP out=GRPVEC; 
var &GROUP;
run;

%*Create a macro variable that contains a list of all levels of the grouping;
%*variable;

data _null_;
 set GRPVEC;
 GRPVEC=
 %do GV=1 %to %eval(&NGRPS-1);
        trim(left(COL&GV))||" "||
 %end;
        trim(left(COL&NGRPS));

 call symput('GRPVEC',GRPVEC);

 run;

%*Determine the number of values in the shift list;
%*This quantity should equal the number of groups;
%*If not, the program will terminate;
%*Algorithm for Counting Number of Items in a List from SAS Technical Support Page (http://support.sas.com/faq/016/FAQ01617.html)
FAQ 1617;

%LET COUNT2=1;
  %LET WORD=%QSCAN(&SHIFTLIST,&COUNT2,%STR( ));
  %DO %WHILE(&WORD NE);
    %LET COUNT2=%EVAL(&COUNT2+1);
    %LET WORD=%QSCAN(&SHIFTLIST,&COUNT2,%STR(  ));
  %END;
  %let COUNT=%eval(&COUNT2-1);

%if &COUNT = &NGRPS %then %do;
%*Loop to compare number of items in the shift list with the number of items;
%*in the group list (&NGRPS);
%*If equal, the program continues processing;

  %do G=1 %to &NGRPS;
    %let SAMPLE=%scan(&GRPVEC,&G," ");

    data SAMPLE&G;
     set &dataset.2;

      if &GROUP="&SAMPLE";

    run;

    options nonotes;

      %macro BOOTSTRAP(BDATASET,OUTBDS,BSAMPLE_SIZE);

        data &OUTBDS;
          do S=1 to &BSAMPLE_SIZE; %*Begin of bootstrap (resampling) loop, S;
             SAMPNUM=S;
             PICKIT=ceil(RANUNI(0)*TOTAL_OBS);
             set &BDATASET point=PICKIT nobs=TOTAL_OBS;
             output;
           end; %*End of bootstrap (resampling) loop, S;
           stop;

         run;

      %mend BOOTSTRAP;

    data POWERCOUNT&G;
     SIGCOUNT=.;
    run;

   %do I=1 %to &ITERATION_NO; %*Begin process of performing statistical testing;

    %BOOTSTRAP(SAMPLE&G,BOOTSAMPLE&G,%eval(&NGRPS*&SAMPLE_SIZE));

    data SHIFT&G;
     set BOOTSAMPLE&G;

      if _n_ < %eval(&SAMPLE_SIZE)+1 then do;
       DELTA&VAR=&VAR;
       NEWGROUP=GROUP||"_"||"1";
      end;
      %do S=2 %to %eval(&COUNT-1);
        else if _n_ < %eval(&S*&SAMPLE_SIZE)+1 then do;
        %let DELTA=%scan(&SHIFTLIST,&S," ");
         DELTA&VAR=&VAR+&DELTA;
         NEWGROUP=GROUP||"_"||"&S";
       end; %*end of second _n_ condition;
      %end; %*end of macro S loop;

     else do; %*Final _n_ condition, default if the other two are not satisfied;
       DELTA&VAR=&VAR+%scan(&SHIFTLIST,&COUNT," ");
       NEWGROUP=GROUP||"_"||"&COUNT";
     end; %*End of final _n_ condition;

     %*Perform k sample comparison of locations for km observations in group &G;
      proc npar1way data=SHIFT&G wilcoxon noprint;
       class NEWGROUP;
        var DELTA&VAR;
       output out=SHIFT&G._OUT;
      run;

     %*Are the results of the test statistically significant at level &ALPHA?;
       data COUNTEVENT;
        set SHIFT&G._OUT;
         if .< P_KW <&ALPHA then do;
         SIGCOUNT=1;
         end; %*End of statistical significance check - Assign indicator value of "1";
	          %*if true;
         else do; %*Not statistically significant - Assign indicator value of "0";
          SIGCOUNT=0;
         end; %*end of loop for not statistically significant loop;

       run;

     %*Concanenate &I-th data set together with others;
       data POWERCOUNT&G;
        set POWERCOUNT&G COUNTEVENT;
       run;

   %end; %*end of &ITERATION_NO loop;

   data POWERCOUNT&G;
    set POWERCOUNT&G;
      if SIGCOUNT ne .;
      keep SIGCOUNT;
   run;

   %*Calculate the power for the Gth group;
   %*km observations tested &NGRPS*&SAMPLE_SIZE times;
   proc freq data=POWERCOUNT&G noprint;
    tables SIGCOUNT/out=POWER&G;
   run;

   %*Save only percentage of indicators =1 (statistically significant events);
   data POWER&G;
    set POWER&G;
    if SIGCOUNT=1;
   run;

   %*Concantenate power for all &ITERATION_NO comparisons within &NGRP groups containing;
   %*&NGRPS*&SAMPLE_SIZE observations;
   data POWER;
    set POWER POWER&G;
   run;

  %end; %*end of G loop;

options notes;

%*Detemine the overall (mean) power;
proc means data=POWER mean noprint;
where PERCENT ge 0;
var PERCENT;
output out=MEANPOWER mean=MEANPOWER;
run;

title1 "The Mahoney-Magel generalization of the Collings-Hamilton approach";
title2 "for power estimation with location shifts %left(%trim(&SHIFTLIST))";
title3 "of %left(%trim(&NGRPS)) groups (%left(%trim(&GRPVEC))) of sample size %left(%trim(&SAMPLE_SIZE))";
title4 "compared at the %left(%trim(&ALPHA))-level of statistical significance";
title6 "(Power based on a bootstrap conducted &ITERATION_NO times)";

proc print data=MEANPOWER label noobs;
var MEANPOWER;
label MEANPOWER="POWER";
run;

%end; %*End of &COUNT = &NGRPS condition;
%else %do; %*Condition if &COUNT ne &NGRPS;
%*produce an error message;

   %put **************ERROR********************;
   %put *                                     *;
   %put * The number of groups does not equal *;
   %put * the number of shifts in the list.   *;
   %put * The program will terminate.         *;
   %put *                                     *;
   %put **************ERROR********************;

%end; %*End of &COUNT ne &NGRPS condition/ERROR message;

%mend kwss;
