%macro nparmcc(dataset,group,control,var,alpha);
***************************************************************;
* A Macro for Performing all Pair-wise Nonparametric          *;
* Against a Common Control Group                              *;
* ------------------------------------------------------------*;
*                                                             *;
* This macro is designed to perform all two-sided pair-wise   *;
* comparisons of the location parameters for n groups (n gt 2)*;
* against a designated control group in a one-way layout.     *;
* The method for the procedure involves a large sample        *;
* approximation. The method of comparison employs ranking all *;
* non-missing observations and averaging the ranks for each   *;
* level of the class variable (GROUP).                        *;
*                                                             *;
* The required inputs for the macro are a SAS data set        *;
* (DATASET) containing a character class variable (GROUP),    *;
* a designated level of the class variable serving as the     *;
* control group (CONTROL), a response variable (VAR), and a   *;
* family-wise error rate (ALPHA).                             *;
*                                                             *;
* For more details on this nonparametric multiple comparison  *;
* procedure, see Hollander and Wolfe s Nonparametric          *;
* Statistical Methods, 1/e pp. 130-131 and Miller s           *;
* Simultaneous Statistical Inference, 2/e pp. 129-138.        *;
*                                                             *;
* Overall Macro Scheme:                                       *;
* --------------------                                        *;
* The macro consists of a body of code containing one embedded*;
* macro (%GROUPS). The embedded macro determines the number of*;
* groups present (NGRPS). If a group in the SAS data set does*;
* not contain at least one response value, it will not be     *;
* included in the analysis. The embedded macro also creates   *;
* one global macro variable that contains the group labels    *;
* (GRPVEC) for the levels of the class variable.             *;
*                                                             *;
* The main body of the SAS macro code determines summary      *;
* statistics (e.g., average ranks, sample sizes, etc.) This   *;
* information is then employed to calculate the pair-wise     *;
* test statistics. The cutoff for the test statistic is       *;
* calculated with PROBIT function. The results are then       *;
* printed out with a PROC PRINT procedure.                    *;
*                                                             *;
***************************************************************;

%macro GROUPS;
%global GRPVEC;

proc sort data=&dataset; by &GROUP;

data &GROUP; set &dataset; by &GROUP;
if first.&GROUP;
if &GROUP ne "&CONTROL";
run;

proc transpose data=&GROUP out=GRPVEC; var &GROUP;

data _null_;
call symput('NGRPS',left(put(count,8.)));
stop;
set &GROUP nobs=COUNT;
run;

data null; set GRPVEC;

length GRPVEC $100;

GRPVEC=
%do G=1 %to %eval(&NGRPS-1);
 trim(left(COL&G))||" "||
%end;
 trim(left(COL&NGRPS));

GRPVEC="&CONTROL"||" "||GRPVEC;

call SYMPUT('GRPVEC',GRPVEC);
run;

%mend GROUPS;

%GROUPS

proc sort data=&dataset;
by &GROUP;

proc rank data=&dataset out=R&VAR;
ranks R&VAR;
var &VAR;
run;

proc sort data=R&VAR;
by &GROUP;
run;

proc univariate noprint data=R&VAR;
var R&VAR;
by &GROUP;
output out=S&VAR sum=S&VAR n=N&VAR;
run;

proc univariate noprint data=S&VAR;
var N&VAR;
output out=SN&VAR sum=SN&VAR;
run;

data _null_;
 set SN&VAR;

call symput("N",left(put(SN&VAR,8.)));
run;

data SR&VAR; set S&VAR;
 if &GROUP ne "&CONTROL";
keep S&VAR;
run;

data _null_; set S&VAR;
 if &GROUP = "&CONTROL";
 call symput("NC",N&VAR);
run;

data N&VAR; set S&VAR;
 if &GROUP ne "&CONTROL";
keep N&VAR;
run;

proc transpose data=SR&VAR out=SRVEC; var S&VAR;
run;

data _null_;
 set S&VAR;
 if &GROUP = "&CONTROL";

 call symput("SRC",left(put(S&VAR,11.6)));

run;

data _null_;
 set SRVEC;

length SRVEC $50;

SRVEC=
%do S=1 %to %eval(&NGRPS-1);
 trim(left(put(COL&S,11.3)))||" "||
%end;
 trim(left(put(COL&NGRPS,11.3)));

call SYMPUT('SRVEC',SRVEC);
run;

%let SRVEC=&SRC &SRVEC;

proc transpose data=N&VAR out=nVEC; var n&VAR;
run;

data _null_;
 set NVEC;

nVEC=
%do S2=1 %to %eval(&NGRPS-1);
 trim(left(put(COL&S2,8.)))||" "||
%end;
 trim(left(put(COL&NGRPS,8.)));

nVEC=&NC||" "||nVEC;

call SYMPUT('NVEC',NVEC);
run;

data NPARMCP;

length DIFFLBL $50;

%do U=2 %to %eval(&NGRPS+1);

 %do V=1 %to 1;

  %let RSUM&U=%scan(&SRVEC,&U," ");
  %let N&U=%scan(&NVEC,&U," ");
  %let RSUM&V=%scan(&SRVEC,&V," ");
  %let N&V=%scan(&NVEC,&V," ");

  c+1;
  DIFFLBL="%scan(&GRPVEC,&U," ") - &CONTROL";
  AVDIFF=abs((&&RSUM&U/&&N&U)-(&&RSUM&V/&&N&V));
  DCUTOFF=probit(1-(&ALPHA/&NGRPS))
         *(((&N*(&N+1))/12)**0.5)
         *((1/&&N&U + 1/&&N&V)**0.5);
  MCUTOFF= probmc('MAXMOD',.,(1-&ALPHA),.,&NGRPS)
         *(((&N*(&N+1))/12)**0.5)
         *((1/&&N&U + 1/&&N&V)**0.5);
  SYMBOLD="    ";
  SYMBOLM="    ";
  if AVDIFF ge DCUTOFF then do;
   SYMBOLD="*D*";
  end;
  if AVDIFF ge MCUTOFF then do;
   SYMBOLM="*M*";
  end;
  output;
 %end;

%end;
label difflbl="Group@comparisons"
      avdiff="Difference@in@average@ranks"
      dcutoff="Dunn@cutoff@at@alpha=&alpha"
      mcutoff="Miller@cutoff@at@alpha=&alpha"
      symbold="Significant@difference@Dunn's = *D*"
      symbolm="Significant@difference@Miller's = *D*"
      c="Comparison@number"
      ;
title1 "Large sample approximation multiple comparison procedure";
title2 "for all treatment compared against a control (control group = &control)";
title3 "%trim(%left(%eval(&ngrps+1))) groups: %trim(%left(&grpvec)) (respective sample sizes: %trim(%left(&nvec)))";
title4 "Alpha = &alpha";
proc print label split='@' noobs;
run;

%mend nparmcc;
